#
# jf_spheretool.rb 2010 jim.foltz@gmail.com
#
# Menu: Draw/Sphere
#
# Version 1.4 (2010, Dec 13) <jim.foltz@gmail.com>
#   * Fixed getExtents method - no more clipping.
#   * Added line thickness when inference locked.
#   * VCB Label changes top Diameter in Diam mode.
#   * Consistent status bar texts.
#
# Version 1.3 
#   * Added diamter mode - press the TAB key to create
#     a sphere from 2 points
#   * also, suffix a radius with 'd' to create a diameter in
#     the Measurements (VCB) box
#
# Version 1.2 
#   * Fixed mis-matched paranthesis that worked on my system
#     but not on another. Wierd.
#   * Changed Sketchup::set_status_text to Sketchup.set_status_text
#
# Version 1.1 
#   Added check for reversed faces
#
# Ripped directly from linetool.rb distributed with Google SketchUp
#
# Copyright 2006, Google Inc.

# This software is provided as an example of using the Ruby interface
# to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------

require 'sketchup.rb'

#-----------------------------------------------------------------------------

# To create a new tool in Ruby, you must define a new class that implements
# the methods for the events that you want to resplond to.  You do not have
# to implement methods for every possible event that a Tool can respond to.

# Once you have defined a tool class, you select that tool by creating an
# instance of it and passsing it to Sketchup.active_model.select_tool

# This implementation of a tool tries to be pretty complete to show all
# of the kinds of things that you can do in a tool.  This makes it a little
# complicated.  You should also look at the TrackMouseTool defined in
# utilities.rb for an example of a simpler tool.

# This example shows the implementation of a simple sphere tool.  This tool
# is similar to the Circle tool in SketchUp

module JF
    class SphereTool

        # This is the standard Ruby initialize method that is called when you create
        # a new object.
        def initialize
            @numsegs = 24
            @ip1 = nil
            @ip2 = nil
            @xdown = 0
            @ydown = 0
        end

        # The activate method is called by SketchUp when the tool is first selected.
        # it is a good place to put most of your initialization
        def activate
            # The Sketchup::InputPoint class is used to get 3D points from screen
            # positions.  It uses the SketchUp inferencing code.
            # In this tool, we will have two points;
            # one for the center, one to calculate the radius
            @ip1 = Sketchup::InputPoint.new
            @ip2 = Sketchup::InputPoint.new
            @ip = Sketchup::InputPoint.new
            @drawn = false
            @diamode = false

            # This sets the label for the VCB
            #Sketchup::set_status_text $exStrings.GetString("Length"), SB_VCB_LABEL
            #Sketchup::set_status_text $exStrings.GetString("Radius"), SB_VCB_LABEL
            Sketchup.set_status_text("Segments", SB_VCB_LABEL)
            Sketchup.set_status_text("#{@numsegs}", SB_VCB_VALUE)

            self.reset(nil)
        end

        # deactivate is called when the tool is deactivated because
        # a different tool was selected
        def deactivate(view)
            view.invalidate if @drawn
        end

        def onMouseMove(flags, x, y, view)
            if( @state == 0 )
                @ip.pick view, x, y
                if( @ip != @ip1 )
                    view.invalidate if( @ip.display? or @ip1.display? )
                    @ip1.copy! @ip
                    view.tooltip = @ip1.tooltip
                end
            else
                @ip2.pick view, x, y, @ip1
                view.tooltip = @ip2.tooltip if( @ip2.valid? )
                view.invalidate

                if( @ip2.valid? )
                    length = @ip1.position.distance(@ip2.position)
                    Sketchup.set_status_text(length.to_s, SB_VCB_VALUE)
                end

                if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
                    @dragging = true
                end
            end
        end

        def resume(view)
            if @state == 0
                Sketchup.set_status_text("Enter segments [#{@numsegs}], or select center point.", SB_PROMPT)
                Sketchup.set_status_text("Segments", SB_VCB_LABEL)
                Sketchup.set_status_text("#{@numsegs}", SB_VCB_VALUE)
            elsif @state == 1
                if @diamode
                    Sketchup.set_status_text("Type diameter, or select 2nd point. Tab=Rad, Esc=Reset.", SB_PROMPT)
                    Sketchup.set_status_text("Diameter", SB_VCB_LABEL)
                else
                    Sketchup.set_status_text("Type radius, or select 2nd point. Tab=Dia, Esc=Reset.", SB_PROMPT)
                    Sketchup.set_status_text("Radius", SB_VCB_LABEL)
                end
            end
        end
        def onLButtonDown(flags, x, y, view)
            if( @state == 0 )
                @ip1.pick view, x, y
                if( @ip1.valid? )
                    @state = 1
                    #Sketchup::set_status_text $exStrings.GetString("Select second end"), SB_PROMPT
                    if @diamode
                        Sketchup.set_status_text("Type diam, or select 2nd point. Tab=Rad, Esc=Reset.", SB_PROMPT)
                        Sketchup.set_status_text("Diam", SB_VCB_LABEL)
                    else
                        Sketchup.set_status_text("Type radius, or select 2nd point. Tab=Dia, Esc=Reset.", SB_PROMPT)
                        Sketchup.set_status_text("Radius", SB_VCB_LABEL)
                    end
                    #Sketchup.set_status_text("Radius", SB_VCB_LABEL)
                    @xdown = x
                    @ydown = y
                end
            else
                # create the sphere on the second click
                if( @ip2.valid? )
                    self.create_geometry(@ip1.position, @ip2.position,view)
                    self.reset(view)
                end
            end

            # Clear any inference lock
            view.lock_inference
        end

        # The onLButtonUp method is called when the user releases the left mouse button.
        def onLButtonUp(flags, x, y, view)
            # If we are doing a drag, then create the line on the mouse up event
            if( @dragging && @ip2.valid? )
                self.create_geometry(@ip1.position, @ip2.position,view)
                self.reset(view)
            end
        end

        def onKeyDown(key, repeat, flags, view)
            if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
                @shift_down_time = Time.now

                # if we already have an inference lock, then unlock it
                if( view.inference_locked? )
                    # calling lock_inference with no arguments actually unlocks
                    view.lock_inference
                elsif( @state == 0 && @ip1.valid? )
                    view.lock_inference @ip1
                elsif( @state == 1 && @ip2.valid? )
                    view.lock_inference @ip2, @ip1
                end
            end
            if key == 9
                @diamode = !@diamode
                if @diamode
                    Sketchup.set_status_text("Type diam, or select 2nd point. Tab=Rad, Esc=Reset.", SB_PROMPT)
                    Sketchup.set_status_text("Diameter", SB_VCB_LABEL)
                else
                    Sketchup.set_status_text("Type radius, or select 2nd point. Tab=Dia, Esc=Reset.", SB_PROMPT)
                    Sketchup.set_status_text("Radius", SB_VCB_LABEL)
                end
            end
        end

        def onKeyUp(key, repeat, flags, view)
            if( key == CONSTRAIN_MODIFIER_KEY && view.inference_locked? && (Time.now - @shift_down_time) > 0.5 )
                view.lock_inference
            end
        end

        def onUserText(text, view)
            if @state == 0
                @numsegs = text.to_l.to_i
                #Sketchup.set_status_text("", SB_VCB_VALUE)
                #Sketchup.set_status_text("Radius", SB_VCB_LABEL)
                #Sketchup.set_status_text("Select center:", SB_PROMPT)
                return 
            end
            return if not @ip2.valid?

            puts "text:#{text.inspect}"
            dia = 1.0
            if text[-1] == ?d
                dia = 2.0
                text.chop!
            end
            begin
                value = text.to_l
            rescue
                # Error parsing the text
                UI.beep
                puts "Cannot convert #{text} to a Length"
                value = nil
                Sketchup.set_status_text("", SB_VCB_VALUE)
            end
            return if !value

            # Compute the direction and the second point
            pt1 = @ip1.position
            vec = @ip2.position - pt1
            if( vec.length == 0.0 )
                UI.beep
                return
            end
            vec.length = value / dia
            pt2 = pt1 + vec

            # Create a line
            self.create_geometry(pt1, pt2, view)
            self.reset(view)
        end

        def draw(view)
            if( @ip1.valid? )
                if( @ip1.display? )
                    @ip1.draw(view)
                    @drawn = true
                end

                if( @ip2.valid? )
                    @ip2.draw(view) if( @ip2.display? )

                    # The set_color_from_line method determines what color
                    # to use to draw a line based on its direction.  For example
                    # red, green or blue.
                    view.set_color_from_line(@ip1, @ip2)
                    self.draw_geometry(@ip1.position, @ip2.position, view)
                    @drawn = true
                end
            end
        end

        # onCancel is called when the user hits the escape key
        def onCancel(flag, view)
            self.reset(view)
        end



        # Reset the tool back to its initial state
        def reset(view)
            @state = 0
            Sketchup.set_status_text("Segments", SB_VCB_LABEL)
            Sketchup.set_status_text("#{@numsegs}", SB_VCB_VALUE)
            Sketchup.set_status_text("Enter segments [#{@numsegs}], or select center point. Tab=Dia Esc=Reset", SB_PROMPT)
            @ip1.clear
            @ip2.clear
            if( view )
                view.tooltip = nil
                view.invalidate if @drawn
            end
            @drawn = false
            @dragging = false
        end

        # Create new geometry when the user has selected two points.
        def create_geometry(p1, p2, view)
            model = view.model
            model.start_operation("Draw Sphere", true)
            radius = p2.distance(p1)
            if @diamode
                radius = p2.distance(p1) / 2.0
                p1 = Geom.linear_combination(0.5, p1, 0.5, p2)
            end
            group = view.model.entities.add_group
            group.name = "Sphere"
            entities = group.entities
            #entities = model.entities
            circle1 = entities.add_circle p1, [0,1,0], radius, @numsegs
            circle2 = entities.add_circle p1+[0,0,-2*radius], [0,0,1], 2.0 * radius, @numsegs
            face1 = entities.add_face circle1
            #face2 = entities.add_face circle2
            face1.followme circle2
            entities.erase_entities circle2
            #view.model.entities.add_cline(p1,p2)
            # Blue Balls? Reverse Faces!
            #ray = Geom::Vector3d.new(0.98252, -0.131633, 0.131633)
            #item = model.raytest([p1, ray])
            #face = item[1][1]
            #normal = face.normal
            #if normal.x < 0
            #    entities.each{|e| e.reverse! if e.respond_to? :reverse!}
            #end
            model.commit_operation
        end

        def sp_points_on_circle(center, normal, radius, numseg, rotAngle)
            # Get the x and y axes
            axes = Geom::Vector3d.new(normal).axes
            center = Geom::Point3d.new(center)
            xaxis = axes[0]
            yaxis = axes[1]

            xaxis.length = radius
            yaxis.length = radius

            if(rotAngle==nil)
                rotAngle=0.0
            end

            # compute the points
            da = (Math::PI * 2) / numseg
            pts = []
            for i in 0...numseg do
                angle = rotAngle + (i * da)
                cosa = Math.cos(angle)
                sina = Math.sin(angle)
                vec = Geom::Vector3d.linear_combination(cosa, xaxis, sina, yaxis)
                pts.push(center + vec)
            end
            # close the circle
            pts.push(pts[0].clone)
            pts
        end

        def draw_geometry(pt1, pt2, view)
            center = pt1.clone
            normal = pt2.vector_to pt1
            normal = Z_AXIS
            radius = pt2.distance pt1
            if @diamode
                #center.offset!(pt2 - pt1)
                center = Geom.linear_combination(0.5, pt1, 0.5, pt2)
                radius /= 2.0
            end
            angleDelta=0
            @shapePoints=sp_points_on_circle(center, normal, radius, @numsegs, angleDelta)
            #@shapePoints = TT::Geom3d.circle(center, normal, radius, @numsegs)
            @shapeRadius=radius
            view.drawing_color="black"
            view.draw(GL_LINE_STRIP, @shapePoints)
            if view.inference_locked?
                view.line_width = 2
            else
                view.line_width = 1
            end
            view.set_color_from_line(pt1, pt2)
            view.draw_line(pt1, pt2)
            view.invalidate
        end
        def getExtents
            if @shapePoints
                sp = @shapePoints.dup
                bb = Geom::BoundingBox.new
                sp.push(sp[0])
                sp.each {|pt| bb.add(pt)}
                return bb
            end
        end
    end # class SphereTool
end

unless file_loaded?("jf_spheretool.rb")
    ($submenu || UI.menu("Draw")).add_item("SphereTool") { Sketchup.active_model.select_tool(JF::SphereTool.new)}
    file_loaded("jf_spheretool.rb")
end
